# __init__.py

__version__ = version = "1.1.1"
__version_tuple__ = version_tuple = (1, 1, 1)
__copyright__ = "Copyright (c) 2024 Tom CHAMBARETAUD"

from resetTolkien.cli import main

if __name__ == "__main__":
    main()
